import click
import questionary as q
import docker

from colorama import Fore, Style

from vantage6.common import error, info, debug
from vantage6.common.docker.addons import check_docker_running


@click.command()
def cli_node_clean() -> None:
    """
    Erase temporary Docker volumes.
    """
    check_docker_running()
    client = docker.from_env()

    # retrieve all volumes
    volumes = client.volumes.list()
    candidates = []
    msg = "This would remove the following volumes: "
    for volume in volumes:
        if volume.name[-6:] == "tmpvol":
            candidates.append(volume)
            msg += volume.name + ","
    info(msg)

    try:
        confirm = q.confirm("Are you sure?").unsafe_ask()
    except KeyboardInterrupt:
        confirm = False

    if confirm:
        for volume in candidates:
            try:
                volume.remove()
                # info(volume.name)
            except docker.errors.APIError as e:
                error(
                    f"Failed to remove volume {Fore.RED}'{volume.name}'"
                    f"{Style.RESET_ALL}. Is it still in use?"
                )
                debug(e)
                exit(1)
    info("Done!")



from pathlib import Path

import questionary as q

from vantage6.common import generate_apikey
from vantage6.common.globals import (
    DATABASE_TYPES,
    InstanceType,
    NodePolicy,
    Ports,
    DEFAULT_API_PATH,
)
from vantage6.common.client.node_client import NodeClient
from vantage6.common.context import AppContext
from vantage6.common import error, warning, info
from vantage6.cli.context import select_context_class
from vantage6.cli.configuration_manager import (
    NodeConfigurationManager,
    ServerConfigurationManager,
)


def node_configuration_questionaire(dirs: dict, instance_name: str) -> dict:
    """
    Questionary to generate a config file for the node instance.

    Parameters
    ----------
    dirs : dict
        Dictionary with the directories of the node instance.
    instance_name : str
        Name of the node instance.

    Returns
    -------
    dict
        Dictionary with the new node configuration
    """
    config = q.unsafe_prompt(
        [
            {"type": "text", "name": "api_key", "message": "Enter given api-key:"},
            {
                "type": "text",
                "name": "server_url",
                "message": "The base-URL of the server:",
                "default": "http://localhost",
            },
        ]
    )
    # remove trailing slash from server_url if entered by user
    config["server_url"] = config["server_url"].rstrip("/")

    # set default port to the https port if server_url is https
    default_port = (
        str(Ports.HTTPS.value)
        if config["server_url"].startswith("https")
        else str(Ports.DEV_SERVER.value)
    )

    config = config | q.unsafe_prompt(
        [
            {
                "type": "text",
                "name": "port",
                "message": "Enter port to which the server listens:",
                "default": default_port,
            },
            {
                "type": "text",
                "name": "api_path",
                "message": "Path of the api:",
                "default": "/api",
            },
            {
                "type": "text",
                "name": "task_dir",
                "message": "Task directory path:",
                "default": str(dirs["data"]),
            },
        ]
    )

    config["databases"] = list()
    while q.confirm("Do you want to add a database?").unsafe_ask():
        db_label = q.unsafe_prompt(
            [
                {
                    "type": "text",
                    "name": "label",
                    "message": "Enter unique label for the database:",
                    "default": "default",
                }
            ]
        )
        db_path = q.unsafe_prompt(
            [{"type": "text", "name": "uri", "message": "Database URI:"}]
        )
        db_type = q.select("Database type:", choices=DATABASE_TYPES).unsafe_ask()

        config["databases"].append(
            {"label": db_label.get("label"), "uri": db_path.get("uri"), "type": db_type}
        )

    is_add_vpn = q.confirm(
        "Do you want to connect to a VPN server?", default=False
    ).unsafe_ask()
    if is_add_vpn:
        config["vpn_subnet"] = q.text(
            message="Subnet of the VPN server you want to connect to:",
            default="10.76.0.0/16",
        ).unsafe_ask()

    is_policies = q.confirm(
        "Do you want to limit the algorithms allowed to run on your node? This "
        "should always be done for production scenarios.",
        default=True,
    ).unsafe_ask()
    policies = {}
    if is_policies:
        info(
            "You can limit the algorithms that can run on your node in two ways: by "
            "allowing specific algorithms or by allowing all algorithms in a given "
            "algorithm store."
        )
        ask_single_algorithms = q.confirm(
            "Do you want to enter a list of allowed algorithms?"
        ).unsafe_ask()
        if ask_single_algorithms:
            policies[NodePolicy.ALLOWED_ALGORITHMS.value] = _get_allowed_algorithms()
        ask_algorithm_stores = q.confirm(
            "Do you want to allow algorithms from specific algorithm stores?"
        ).unsafe_ask()
        if ask_algorithm_stores:
            policies[NodePolicy.ALLOWED_ALGORITHM_STORES.value] = (
                _get_allowed_algorithm_stores()
            )
        if ask_single_algorithms and ask_algorithm_stores:
            require_both_whitelists = q.confirm(
                "Do you want to allow only algorithms that are both in the list of "
                "allowed algorithms *AND* are part of one of the allowed algorithm "
                "stores? If not, algorithms will be allowed if they are in either the "
                "list of allowed algorithms or one of the allowed algorithm stores.",
                default=True,
            ).unsafe_ask()
            policies["allow_either_whitelist_or_store"] = not require_both_whitelists
    if policies:
        config["policies"] = policies

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
    ).unsafe_ask()

    config["logging"] = {
        "level": res,
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "requests", "level": "warning"},
            {"name": "engineio.client", "level": "warning"},
            {"name": "docker.utils.config", "level": "warning"},
            {"name": "docker.auth", "level": "warning"},
        ],
    }

    # Check if we can login to the server to retrieve collaboration settings
    client = NodeClient(config["server_url"], config["port"], config["api_path"])
    try:
        client.authenticate(config["api_key"])
    except Exception as e:
        error(f"Could not authenticate with server: {e}")
        error("Please check (1) your API key and (2) if your server is online")
        warning(
            "If you continue, you should provide your collaboration "
            "settings manually."
        )
        if q.confirm("Do you want to abort?", default=True).unsafe_ask():
            exit(0)

    if client.whoami is not None:
        encryption = client.is_encrypted_collaboration()
        # TODO when we build collaboration policies, update this to provide
        # the node admin with a list of all policies, and whether or not
        # to accept them
        q.confirm(
            f"Encryption is {'enabled' if encryption else 'disabled'}"
            f" for this collaboration. Accept?",
            default=True,
        ).unsafe_ask()
    else:
        encryption = q.confirm("Enable encryption?", default=True).unsafe_ask()

    private_key = (
        "" if not encryption else q.text("Path to private key file:").unsafe_ask()
    )

    config["encryption"] = {
        "enabled": encryption is True or encryption == "true",
        "private_key": private_key,
    }

    return config


def _get_allowed_algorithms() -> list[str]:
    """
    Prompt the user for the allowed algorithms on their node

    Returns
    -------
    list[str]
        List of allowed algorithms or regular expressions to match them
    """
    info("Below you can add algorithms that are allowed to run on your node.")
    info(
        "You can use regular expressions to match multiple algorithms, or you can "
        "use strings to provide one algorithm at a time."
    )
    info("Examples:")
    info(r"^harbor2\.vantage6\.ai/demo/average$    Allow the demo average algorithm")
    info(
        r"^harbor2\.vantage6\.ai/algorithms/.*   Allow all algorithms from "
        "harbor2.vantage6.ai/algorithms"
    )
    info(
        r"^harbor2\.vantage6\.ai/demo/average@sha256:82becede...$    Allow a "
        "specific hash of average algorithm"
    )
    allowed_algorithms = []
    while True:
        algo = q.text(message="Enter your algorithm expression:").unsafe_ask()
        allowed_algorithms.append(algo)
        if not q.confirm(
            "Do you want to add another algorithm expression?", default=True
        ).unsafe_ask():
            break
    return allowed_algorithms


def _get_allowed_algorithm_stores() -> list[str]:
    """
    Prompt the user for the allowed algorithm stores on their node

    Returns
    -------
    list[str]
        List of allowed algorithm stores
    """
    info("Below you can add algorithm stores that are allowed to run on your node.")
    info(
        "You can use regular expressions to match multiple algorithm stores, or you can"
        " use strings to provide one algorithm store at a time."
    )
    info("Examples:")
    info(
        "https://store.cotopaxi.vantage6.ai    Allow all algorithms from the "
        "community store"
    )
    info(
        r"^https://*\.vantage6\.ai$               Allow all algorithms from any "
        "store hosted on vantage6.ai"
    )
    allowed_algorithm_stores = []
    while True:
        store = q.text(message="Enter the URL of the algorithm store:").unsafe_ask()
        allowed_algorithm_stores.append(store)
        if not q.confirm(
            "Do you want to add another algorithm store?", default=True
        ).unsafe_ask():
            break
    return allowed_algorithm_stores


def _get_common_server_config(instance_type: InstanceType, instance_name: str) -> dict:
    """
    Part of the questionaire that is common to all server types (vantage6
    server and algorithm store server).

    Parameters
    ----------
    instance_type : InstanceType
        Type of server instance.
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with new (partial) server configuration
    """
    config = q.unsafe_prompt(
        [
            {
                "type": "text",
                "name": "description",
                "message": "Enter a human-readable description:",
            },
            {"type": "text", "name": "ip", "message": "ip:", "default": "0.0.0.0"},
            {
                "type": "text",
                "name": "port",
                "message": "Enter port to which the server listens:",
                "default": (
                    # Note that .value is required in YAML to get proper str value
                    str(Ports.DEV_SERVER.value)
                    if instance_type == InstanceType.SERVER
                    else str(Ports.DEV_ALGO_STORE.value)
                ),
            },
        ]
    )

    config.update(
        q.unsafe_prompt(
            [
                {
                    "type": "text",
                    "name": "uri",
                    "message": "Database URI:",
                    "default": "sqlite:///default.sqlite",
                },
                {
                    "type": "select",
                    "name": "allow_drop_all",
                    "message": "Allowed to drop all tables: ",
                    "choices": ["True", "False"],
                },
            ]
        )
    )

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
    ).unsafe_ask()

    config["logging"] = {
        "level": res,
        "file": f"{instance_name}.log",
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "socketIO-client", "level": "warning"},
            {"name": "socketio.server", "level": "warning"},
            {"name": "engineio.server", "level": "warning"},
            {"name": "sqlalchemy.engine", "level": "warning"},
            {"name": "requests_oauthlib.oauth2_session", "level": "warning"},
        ],
    }

    config["api_path"] = DEFAULT_API_PATH

    return config


def server_configuration_questionaire(instance_name: str) -> dict:
    """
    Questionary to generate a config file for the server instance.

    Parameters
    ----------
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with the new server configuration
    """

    config = _get_common_server_config(InstanceType.SERVER, instance_name)

    constant_jwt_secret = q.confirm("Do you want a constant JWT secret?").unsafe_ask()
    if constant_jwt_secret:
        config["jwt_secret_key"] = generate_apikey()

    is_mfa = q.confirm("Do you want to enforce two-factor authentication?").unsafe_ask()
    if is_mfa:
        config["two_factor_auth"] = is_mfa

    current_server_url = f"http://localhost:{config['port']}{config['api_path']}"
    config["server_url"] = q.text(
        "What is the server url exposed to the users? If you are running a"
        " development server running locally, this is the same as the "
        "server url. If you are running a production server, this is the "
        "url that users will connect to.",
        default=current_server_url,
    ).unsafe_ask()

    is_add_vpn = q.confirm(
        "Do you want to add a VPN server?", default=False
    ).unsafe_ask()
    if is_add_vpn:
        vpn_config = q.unsafe_prompt(
            [
                {
                    "type": "text",
                    "name": "url",
                    "message": "VPN server URL:",
                },
                {
                    "type": "text",
                    "name": "portal_username",
                    "message": "VPN portal username:",
                },
                {
                    "type": "password",
                    "name": "portal_userpass",
                    "message": "VPN portal password:",
                },
                {
                    "type": "text",
                    "name": "client_id",
                    "message": "VPN client username:",
                },
                {
                    "type": "password",
                    "name": "client_secret",
                    "message": "VPN client password:",
                },
                {
                    "type": "text",
                    "name": "redirect_url",
                    "message": "Redirect url (should be local address of server)",
                    "default": "http://localhost",
                },
            ]
        )
        config["vpn_server"] = vpn_config

    is_add_rabbitmq = q.confirm(
        "Do you want to add a RabbitMQ message queue?"
    ).unsafe_ask()
    if is_add_rabbitmq:
        rabbit_uri = q.text(message="Enter the URI for your RabbitMQ:").unsafe_ask()
        run_rabbit_locally = q.confirm(
            "Do you want to run RabbitMQ locally? (Use only for testing)"
        ).unsafe_ask()
        config["rabbitmq"] = {
            "uri": rabbit_uri,
            "start_with_server": run_rabbit_locally,
        }

    # add algorithm stores to this server
    is_add_community_store = q.confirm(
        "Do you want to make the algorithms from the community algorithm store "
        "available to your users?"
    ).unsafe_ask()
    algorithm_stores = []
    if is_add_community_store:
        algorithm_stores.append(
            {"name": "Community store", "url": "https://store.cotopaxi.vantage6.ai"}
        )
    add_more_stores = q.confirm(
        "Do you want to add more algorithm stores?", default=False
    ).unsafe_ask()
    while add_more_stores:
        store_name = q.text(message="Enter the name of the store:").unsafe_ask()
        store_url = q.text(message="Enter the URL of the store:").unsafe_ask()
        algorithm_stores.append({"name": store_name, "url": store_url})
        add_more_stores = q.confirm(
            "Do you want to add more algorithm stores?", default=False
        ).unsafe_ask()
    config["algorithm_stores"] = algorithm_stores

    return config


def algo_store_configuration_questionaire(instance_name: str) -> dict:
    """
    Questionary to generate a config file for the algorithm store server
    instance.

    Parameters
    ----------
    instance_name : str
        Name of the server instance.

    Returns
    -------
    dict
        Dictionary with the new server configuration
    """
    config = _get_common_server_config(InstanceType.ALGORITHM_STORE, instance_name)

    default_v6_server_uri = (
        f"http://localhost:{Ports.DEV_SERVER.value}{DEFAULT_API_PATH}"
    )
    default_root_username = "root"

    v6_server_uri = q.text(
        "What is the Vantage6 server linked to the algorithm store? "
        "Provide the link to the server endpoint.",
        default=default_v6_server_uri,
    ).unsafe_ask()

    root_username = q.text(
        "What is the username of the root user?",
        default=default_root_username,
    ).unsafe_ask()

    config["root_user"] = {
        "v6_server_uri": v6_server_uri,
        "username": root_username,
    }

    # ask about openness of the algorithm store
    config["policies"] = {"allow_localhost": False}
    is_open = q.confirm(
        "Do you want to open the algorithm store to the public? This will allow anyone "
        "to view the algorithms, but they cannot modify them.",
        default=False,
    ).unsafe_ask()
    if is_open:
        open_algos_policy = "public"
    else:
        is_open_to_whitelist = q.confirm(
            "Do you want to allow all users of whitelisted vantage6 servers to access "
            "the algorithms in the store? If not allowing this, you will have to assign"
            " the appropriate permissions to each user individually.",
            default=True,
        ).unsafe_ask()
        open_algos_policy = "whitelisted" if is_open_to_whitelist else "private"
    config["policies"]["algorithm_view"] = open_algos_policy

    return config


def configuration_wizard(
    type_: InstanceType, instance_name: str, system_folders: bool
) -> Path:
    """
    Create a configuration file for a node or server instance.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    instance_name : str
        Name of the instance
    system_folders : bool
        Whether to use the system folders or not

    Returns
    -------
    Path
        Path to the configuration file
    """
    # for defaults and where to save the config
    dirs = AppContext.instance_folders(type_, instance_name, system_folders)

    # invoke questionaire to create configuration file
    if type_ == InstanceType.NODE:
        conf_manager = NodeConfigurationManager
        config = node_configuration_questionaire(dirs, instance_name)
    elif type_ == InstanceType.SERVER:
        conf_manager = ServerConfigurationManager
        config = server_configuration_questionaire(instance_name)
    else:
        conf_manager = ServerConfigurationManager
        config = algo_store_configuration_questionaire(instance_name)

    # in the case of an environment we need to add it to the current
    # configuration. In the case of application we can simply overwrite this
    # key (although there might be environments present)
    config_file = Path(dirs.get("config")) / (instance_name + ".yaml")

    if Path(config_file).exists():
        config_manager = conf_manager.from_file(config_file)
    else:
        config_manager = conf_manager(instance_name)

    config_manager.put(config)
    config_manager.save(config_file)

    return config_file


def select_configuration_questionaire(type_: InstanceType, system_folders: bool) -> str:
    """
    Ask which configuration the user wants to use. It shows only configurations
    that are in the default folder.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    system_folders : bool
        Whether to use the system folders or not

    Returns
    -------
    str
        Name of the configuration
    """
    context = select_context_class(type_)
    configs, _ = context.available_configurations(system_folders)

    # each collection (file) can contain multiple configs. (e.g. test,
    # dev)
    choices = []
    for config_collection in configs:
        choices.append(
            q.Choice(title=f"{config_collection.name:25}", value=config_collection.name)
        )

    if not choices:
        raise Exception("No configurations could be found!")

    # pop the question
    try:
        return q.select(
            "Select the configuration you want to use:", choices=choices
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)



from __future__ import annotations
from pathlib import Path

from vantage6.common.globals import APPNAME, InstanceType
from vantage6.cli.configuration_manager import ServerConfigurationManager
from vantage6.cli.globals import (
    DEFAULT_SERVER_SYSTEM_FOLDERS as S_FOL,
    PROMETHEUS_DIR,
    ServerType,
    ServerGlobals,
)
from vantage6.cli._version import __version__
from vantage6.cli.context.base_server import BaseServerContext


class ServerContext(BaseServerContext):
    """
    Server context

    Parameters
    ----------
    instance_name : str
        Name of the configuration instance, corresponds to the filename
        of the configuration file.
    system_folders : bool, optional
        System wide or user configuration, by default S_FOL
    """

    # The server configuration manager is aware of the structure of the server
    # configuration file and makes sure only valid configuration can be loaded.
    INST_CONFIG_MANAGER = ServerConfigurationManager

    def __init__(self, instance_name: str, system_folders: bool = S_FOL):
        super().__init__(
            InstanceType.SERVER, instance_name, system_folders=system_folders
        )
        self.log.info("vantage6 version '%s'", __version__)

    def get_database_uri(self) -> str:
        """
        Obtain the database uri from the environment or the configuration.

        Returns
        -------
        str
            string representation of the database uri
        """
        return super().get_database_uri(ServerGlobals.DB_URI_ENV_VAR)

    @property
    def docker_container_name(self) -> str:
        """
        Name of the docker container that the server is running in.

        Returns
        -------
        str
            Server's docker container name
        """
        return f"{APPNAME}-{self.name}-{self.scope}-{ServerType.V6SERVER}"

    @property
    def prometheus_container_name(self) -> str:
        """
        Get the name of the Prometheus Docker container for this server.

        Returns
        -------
        str
            Prometheus container name, unique to this server instance.
        """
        return f"{APPNAME}-prometheus"

    @property
    def prometheus_dir(self) -> Path:
        """
        Get the Prometheus directory path.

        Returns
        -------
        Path
            Path to the Prometheus directory
        """
        return self.data_dir / PROMETHEUS_DIR

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = S_FOL
    ) -> ServerContext:
        """
        Create a server context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        ServerContext
            Server context object
        """
        return super().from_external_config_file(
            path, ServerType.V6SERVER, ServerGlobals.CONFIG_NAME_ENV_VAR, system_folders
        )

    @classmethod
    def config_exists(cls, instance_name: str, system_folders: bool = S_FOL) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().config_exists(
            InstanceType.SERVER, instance_name, system_folders=system_folders
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = S_FOL
    ) -> tuple[list, list]:
        """
        Find all available server configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(InstanceType.SERVER, system_folders)



import time
from threading import Thread

import click
import questionary as q
import docker

from colorama import Fore, Style

from vantage6.common import info, error
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME, InstanceType

from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli.common.utils import print_log_worker


@click.command()
@click.option("-n", "--name", default=None, help="configuration name")
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user", "system_folders", flag_value=False, default=DEFAULT_SERVER_SYSTEM_FOLDERS
)
def cli_server_attach(name: str, system_folders: bool) -> None:
    """
    Show the server logs in the current console.
    """
    check_docker_running()
    client = docker.from_env()

    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.SERVER.value}"}
    )
    running_server_names = [node.name for node in running_servers]

    if not name:
        try:
            name = q.select(
                "Select the server you wish to attach:", choices=running_server_names
            ).unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            return
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}-{InstanceType.SERVER.value}"

    if name in running_server_names:
        container = client.containers.get(name)
        logs = container.attach(stream=True, logs=True, stdout=True)
        Thread(target=print_log_worker, args=(logs,), daemon=True).start()
        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                info("Closing log file. Keyboard Interrupt.")
                info(
                    "Note that your server is still running! Shut it down "
                    f"with {Fore.RED}v6 server stop{Style.RESET_ALL}"
                )
                exit(0)
    else:
        error(f"{Fore.RED}{name}{Style.RESET_ALL} was not running!?")



RABBITMQ_DEFINITIONS = {
    "rabbit_version": "3.6.6",
    "users": [
        {
            "name": "{{username}}",
            "password_hash": "{{password}}",
            "hashing_algorithm": "rabbit_password_hashing_sha256",
            "tags": "administrator",
        }
    ],
    "vhosts": [{"name": "{{vhost_name}}"}],
    "permissions": [
        {
            "user": "{{username}}",
            "vhost": "{{vhost_name}}",
            "configure": ".*",
            "write": ".*",
            "read": ".*",
        }
    ],
    "parameters": [],
    "policies": [],
    "queues": [],
    "exchanges": [],
    "bindings": [],
}



"""
The context module in the CLI package contains the Context classes of instances
started from the CLI, such as nodes and servers. These contexts are related to
the host system and therefore part of the CLI package.

All classes are derived from the abstract AppContext class and provide the
vantage6 applications with naming conventions, standard file locations, and
more.
"""

from colorama import Fore, Style

from vantage6.common.globals import InstanceType
from vantage6.common import error
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.context.server import ServerContext


def select_context_class(
    type_: InstanceType,
) -> ServerContext | NodeContext | AlgorithmStoreContext:
    """
    Select the context class based on the type of instance.

    Parameters
    ----------
    type_ : InstanceType
        The type of instance for which the context should be inserted

    Returns
    -------
    ServerContext | NodeContext | AlgorithmStoreContext
        Specialized subclass of AppContext for the given instance type

    Raises
    ------
    NotImplementedError
        If the type_ is not implemented
    """
    if type_ == InstanceType.SERVER:
        return ServerContext
    elif type_ == InstanceType.ALGORITHM_STORE:
        return AlgorithmStoreContext
    elif type_ == InstanceType.NODE:
        return NodeContext
    else:
        raise NotImplementedError


def get_context(
    type_: InstanceType, name: str, system_folders: bool
) -> ServerContext | NodeContext | AlgorithmStoreContext:
    """
    Load the server context from the configuration file.

    Parameters
    ----------
    type_ : InstanceType
        The type of instance to get the context for
    name : str
        Name of the instance
    system_folders : bool
        Wether to use system folders or if False, the user folders

    Returns
    -------
    AppContext
        Specialized subclass context of AppContext for the given instance type
    """
    ctx_class = select_context_class(type_)
    if not ctx_class.config_exists(name, system_folders):
        scope = "system" if system_folders else "user"
        error(
            f"Configuration {Fore.RED}{name}{Style.RESET_ALL} does not "
            f"exist in the {Fore.RED}{scope}{Style.RESET_ALL} folders!"
        )
        exit(1)

    # We do not want to log this here, we do this in the container and not on
    # the host. We only want CLI logging here.
    ctx_class.LOGGING_ENABLED = False

    # create server context, and initialize db
    ctx = ctx_class(name, system_folders=system_folders)

    return ctx



import click
import questionary as q
import docker
from colorama import Fore, Style
from docker.client import DockerClient

from vantage6.common import info, warning, error
from vantage6.common.docker.addons import (
    check_docker_running,
    remove_container,
    get_server_config_name,
    get_container,
    get_num_nonempty_networks,
    get_network,
    delete_network,
    remove_container_if_exists,
)
from vantage6.common.globals import APPNAME, InstanceType
from vantage6.common import split_rabbitmq_uri
from vantage6.cli.context.server import ServerContext
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli.server.common import get_server_context, stop_ui


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user", "system_folders", flag_value=False, default=DEFAULT_SERVER_SYSTEM_FOLDERS
)
@click.option("--all", "all_servers", flag_value=True, help="Stop all servers")
def cli_server_stop(name: str, system_folders: bool, all_servers: bool):
    """
    Stop one or all running server(s).
    """
    check_docker_running()
    client = docker.from_env()

    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type={InstanceType.SERVER.value}"}
    )

    if not running_servers:
        warning("No servers are currently running.")
        return

    running_server_names = [server.name for server in running_servers]

    if all_servers:
        for container_name in running_server_names:
            _stop_server_containers(client, container_name, system_folders)
        return

    # make sure we have a configuration name to work with
    if not name:
        try:
            container_name = q.select(
                "Select the server you wish to stop:", choices=running_server_names
            ).unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            return
    else:
        post_fix = "system" if system_folders else "user"
        container_name = f"{APPNAME}-{name}-{post_fix}-{InstanceType.SERVER.value}"

    if container_name not in running_server_names:
        error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running!")
        return

    _stop_server_containers(client, container_name, system_folders)


def _stop_server_containers(
    client: DockerClient, container_name: str, system_folders: bool
) -> None:
    """
    Given a server's name, kill its docker container and related (RabbitMQ)
    containers.

    Parameters
    ----------
    client : DockerClient
        Docker client
    container_name : str
        Name of the server to stop
    system_folders : bool
        Wether to use system folders or not
    """
    # kill the server
    remove_container_if_exists(client, name=container_name)
    info(f"Stopped the {Fore.GREEN}{container_name}{Style.RESET_ALL} server.")

    # find the configuration name from the docker container name
    scope = "system" if system_folders else "user"
    config_name = get_server_config_name(container_name, scope)

    ctx = get_server_context(config_name, system_folders, ServerContext)

    # kill the UI container (if it exists)
    stop_ui(client, ctx)

    # delete the server network
    network_name = f"{APPNAME}-{ctx.name}-{ctx.scope}-network"
    network = get_network(client, name=network_name)
    delete_network(network, kill_containers=False)

    # kill RabbitMQ if it exists and no other servers are using to it (i.e. it
    # is not in other docker networks with other containers)
    rabbit_uri = ctx.config.get("rabbitmq", {}).get("uri")
    if rabbit_uri:
        rabbit_container_name = split_rabbitmq_uri(rabbit_uri=rabbit_uri)["host"]
        rabbit_container = get_container(client, name=rabbit_container_name)
        if rabbit_container and get_num_nonempty_networks(rabbit_container) == 0:
            remove_container(rabbit_container, kill=True)
            info(
                f"Stopped the {Fore.GREEN}{rabbit_container_name}"
                f"{Style.RESET_ALL} container."
            )

    if ctx.config.get("prometheus", {}).get("enabled"):
        remove_container_if_exists(client, name=ctx.prometheus_container_name)
        info(
            f"Stopped the {Fore.GREEN}{ctx.prometheus_container_name}"
            f"{Style.RESET_ALL} container."
        )



import subprocess
import click

from vantage6.common import info
from vantage6.client import Client
from vantage6.cli.globals import COMMUNITY_STORE
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.server.start import cli_server_start
from vantage6.cli.dev.utils import get_dev_server_context
from vantage6.common.globals import DEFAULT_API_PATH


@click.command()
@click.option("-n", "--name", default=None, help="Name of the configuration.")
@click.option(
    "-c",
    "--config",
    default=None,
    help="Path to configuration-file; overrides --name",
)
@click.option(
    "--server-image", type=str, default=None, help="Server Docker image to use"
)
@click.option("--node-image", type=str, default=None, help="Node Docker image to use")
@click.option(
    "--store-image", type=str, default=None, help="Algorithm Store Docker image to use"
)
@click.pass_context
def start_demo_network(
    click_ctx: click.Context,
    name: str | None,
    config: str | None,
    server_image: str | None,
    node_image: str | None,
    store_image: str | None,
) -> None:
    """Starts running a demo-network.

    Select a server configuration to run its demo network. You should choose a
    server configuration that you created earlier for a demo network. If you
    have not created a demo network, you can run `v6 dev create-demo-network` to
    create one.
    """
    ctx = get_dev_server_context(config, name)

    # run the server
    info("Starting server...")
    click_ctx.invoke(
        cli_server_start,
        ctx=ctx,
        ip=None,
        port=None,
        image=server_image,
        start_ui=True,
        ui_port=None,
        start_rabbitmq=False,
        rabbitmq_image=None,
        keep=True,
        mount_src="",
        attach=False,
        system_folders=False,
    )

    # run the store
    info("Starting algorithm store...")
    cmd = ["v6", "algorithm-store", "start", "--name", f"{ctx.name}_store", "--user"]
    if store_image:
        cmd.extend(["--image", store_image])
    subprocess.run(cmd, check=True)

    # run all nodes that belong to this server
    configs, _ = NodeContext.available_configurations(system_folders=False)
    node_names = [
        config.name for config in configs if config.name.startswith(f"{ctx.name}_node_")
    ]
    for name in node_names:
        cmd = ["v6", "node", "start", "--name", name]
        if node_image:
            cmd.extend(["--image", node_image])
        subprocess.run(cmd, check=True)

    # now that both server and store have been started, couple them
    info("Linking local algorithm store to server...")
    store_ctxs, _ = AlgorithmStoreContext.available_configurations(system_folders=False)
    store_ctx = [c for c in store_ctxs if c.name == f"{ctx.name}_store"][0]
    client = Client(
        "http://localhost",
        ctx.config["port"],
        ctx.config["api_path"],
        log_level="warn",
    )
    # TODO these credentials are hardcoded and may change if changed elsewhere. Link
    # them together so that they are guaranteed to be the same.
    USERNAME = "dev_admin"
    PASSWORD = "password"
    client.authenticate(USERNAME, PASSWORD)
    existing_stores = client.store.list().get("data", [])
    existing_urls = [store["url"] for store in existing_stores]
    api_path = store_ctx.config.get("api_path", DEFAULT_API_PATH)
    local_store_url = f"http://localhost:{store_ctx.config['port']}{api_path}"
    if not local_store_url in existing_urls:
        client.store.create(
            algorithm_store_url=local_store_url,
            name="local store",
            all_collaborations=True,
            force=True,  # required to link localhost store
        )
        # note that we do not need to register the user as root of the store: this is
        # already handled in the store config file and is executed on store startup (and
        # successful because server is already started up at that point)
    info("Done!")

    # link the community store also to the server
    info("Linking community algorithm store to local server...")
    if not COMMUNITY_STORE in existing_urls:
        client.store.create(
            algorithm_store_url=COMMUNITY_STORE,
            name="Community store (read-only)",
            all_collaborations=True,
            force=True,  # required to continue when linking localhost server
        )
    info("Done!")
